/*******************************************************************************
 * Copyright 2012 pw999
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package net.sf.oqt.core;

import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;

import net.sf.oqt.driver.NullDriver;
import net.sf.oqt.jpa.EntityFinder;

import org.apache.openjpa.persistence.EntityManagerFactoryImpl;
import org.apache.openjpa.persistence.PersistenceProviderImpl;
import org.apache.openjpa.persistence.PersistenceUnitInfoImpl;

/**
 * The CoreFactory centralizes all interactions.
 * <p>
 * The CoreFactory will be responsible for setting up the real connection to the database and the intialization of the
 * EntityManager. It will also be the central reference for the NullDriver and NullConnection since they will add every
 * query they receive to the list maintained by the CoreFactory.
 * <p>
 * Bear in mind that the CoreFactory is not intended to be thread-safe.
 * 
 * @author phillip
 * 
 */
public class CoreFactory {
    /**
     * Contains some general properties which the whole application can access.
     */
    private static final CoreProperties PROPERTIES = new CoreProperties();
    /**
     * The OpenJPA EntityManager. <br />
     * Initialized using lazy loading.
     */
    private static volatile EntityManager EM = null;
    /**
     * All queries executed via the NullDriver will be added here using the
     * NullPreparedStatement and NullCallableStatement.
     */
    private static final List<String> QUERIES = new ArrayList<String>();
    /**
     * A reference to the actual driver. <br />
     * Initialized using lazy loading.
     */
    private static volatile Driver REALDRIVER = null;
    /**
     * A reference to the actual connection.<br />
     * Initialized using lazy loading.
     */
    private static volatile Connection REALCONNECTION = null;

    /**
     * Returns the properties containing all the settings for this application.
     * 
     * @return the properties.
     */
    public static final CoreProperties getProperties() {
        return PROPERTIES;
    }

    /**
     * Returns a reference to the JPA EntityManager.
     * <p>
     * At first call, the EntityManager will be initialized. In order to correctly initialize, the CoreProperties must
     * contains at least the following information:
     * <ul>
     * <li>DriverName
     * <li>URL
     * <li>Username (if required for the connection)
     * <li>Password (if required for the connection)
     * <li>fqns (to get the list of fqns you can use the {@link EntityFinder})
     * </ul>
     * <p>
     * Please bare in mind that the EntityManager will make a connection using the NullDriver, it will never be able to
     * execute statements on the database. It does require a database connection to gather information about your
     * database's capabilities, this connection will be set read-only. The databse itself may be emtpy.
     * <p>
     * It is important to close the database connection afterwards using the {@link #shutdown()} method.
     * 
     * @return the entity manager connected using the NullDriver
     * @throws ClassNotFoundException if the databse driver is not found on the classpath
     * @throws InstantiationException if failed to initliaze the database driver
     * @throws IllegalAccessException if failed to initliaze the database driver
     * @throws SQLException if failed to connect to the database or failed to set the connection to read-only mode.
     */
    public static EntityManager getEntityManager() throws ClassNotFoundException, InstantiationException, IllegalAccessException, SQLException {
        if (EM != null) {
            return EM;
        }
        Class.forName(NullDriver.class.getName()); // register driver for
                                                   // openJPA 2
        initRealDriverAndConnection();

        final Map<String, String> props = new HashMap<String, String>();
        props.put("openjpa.ConnectionURL", "jdbc:tmp/tmp;");
        props.put("openjpa.ConnectionDriverName", NullDriver.class.getName());
        props.put("openjpa.ConnectionUserName", "");
        props.put("openjpa.ConnectionPassword", "");
        // props.put("openjpa.jdbc.SynchronizeMappings", SynchronizeMappings);
        props.put("openjpa.Log", "DefaultLevel=ERROR, Runtime=ERROR, Tool=ERROR, SQL=ERROR, Query=ERROR");
        props.put("openjpa.RuntimeUnenhancedClasses", "supported");
        props.put("openjpa.jdbc.DBDictionary", PROPERTIES.getDictionary());
        props.put("javax.persistence.provider", "org.apache.openjpa.persistence.PersistenceProviderImpl");
        final StringBuilder builder = new StringBuilder();
        builder.append("jpa(Types=");
        Iterator<String> it = PROPERTIES.getFqns().iterator();
        while (it.hasNext()) {
            builder.append(it.next());
            if (it.hasNext()) {
                builder.append(";");
            }
        }
        builder.append(")");
        props.put("openjpa.MetaDataFactory", builder.toString());
        // bypass generic Persistence because it'll try to load the persistence-unit from persistence.xml
        PersistenceProviderImpl provider = new PersistenceProviderImpl();
        PersistenceUnitInfoImpl pui = new PersistenceUnitInfoImpl();
        pui.fromUserProperties(props);
        EntityManagerFactory emf = provider.createContainerEntityManagerFactory(pui, props);
        EM = emf.createEntityManager();
        getVersion(emf);
        return EM;
    }

    /**
     * Closes the entity manager and databse connection.
     * 
     * @throws SQLException if failed to close the connection.
     */
    public static void shutdown() throws SQLException {
        if (REALCONNECTION != null) {
            REALCONNECTION.close();
        }
        if (EM != null) {
            EM.close();
        }
        QUERIES.clear();
        EM = null;
        REALCONNECTION = null;
        REALDRIVER = null;
    }

    static void initRealDriverAndConnection() throws SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException {
        CoreFactory.REALDRIVER = (Driver) Class.forName(PROPERTIES.getDriverName()).newInstance();

        final Properties p = new Properties();
        p.setProperty("user", PROPERTIES.getUsername());
        p.setProperty("password", PROPERTIES.getPassword());
        CoreFactory.REALCONNECTION = REALDRIVER.connect(PROPERTIES.getUrl(), p);
        CoreFactory.REALCONNECTION.setReadOnly(true);
    }

    public static Connection getRealConnection() {
        return REALCONNECTION;
    }

    public static Driver getRealDriver() {
        return REALDRIVER;
    }

    public static List<String> getQueries() {
        return QUERIES;
    }

    @SuppressWarnings("rawtypes")
    static void getVersion(final EntityManagerFactory emf) {
        try {
            Method m = EntityManagerFactoryImpl.class.getMethod("getProperties");
            Map map = (Map) m.invoke(emf);
            PROPERTIES.setJpaVersion((String) map.get("VersionNumber"));
            System.out.println("Running openJPA version " + PROPERTIES.getJpaVersion());
        } catch (Throwable e) {
            e.printStackTrace();
        }
    }

}
