/*******************************************************************************
 * Copyright 2012 pw999
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package net.sf.oqt.jpa;

import java.lang.annotation.Annotation;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import javax.persistence.Entity;
import javax.persistence.MappedSuperclass;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;

import net.sf.oqt.model.EntityVO;
import net.sf.oqt.model.PackageVO;
import net.sf.oqt.model.QueryVO;

import org.reflections.Reflections;


/**
 * The EntityFinder class centralizes all the reflection logic for finding all
 * the entities and all the named queries.
 * 
 * @author phillip
 * 
 */
public class EntityFinder {

    /**
     * This method will find all the entities and named queries based on the list of provided package names.
     * <p>
     * This method uses reflection to find all objects annotated with {@link Entity} or {@link MappedSuperclass}. It
     * will only look in the packages and subpackages which are provided in the basePackages parameter.<br />
     * This method will also extract all the named queries (both from {@link NamedQuery} and {@link NamedQueries}).
     * 
     * @param basePackages the fully qualified names of the packages where to look for the entities.
     * @return an filled in set of PackageVO's
     * @throws ClassNotFoundException this should never occur
     */
    public static Set<PackageVO> findAllEntities(final Collection<String> basePackages) throws ClassNotFoundException {
        final Set<PackageVO> packages = new TreeSet<PackageVO>();
        final Set<Class<?>> classes = findAllEntitiesWithReflection(basePackages);
        for (final Class<?> clazz : classes) {
            final String packageName = clazz.getPackage().getName();
            PackageVO p = new PackageVO(packageName);
            if (packages.contains(p)) {
                final Iterator<PackageVO> it = packages.iterator();
                boolean found = false;
                while (it.hasNext() && !found) {
                    final PackageVO packageInSet = it.next();
                    if (packageInSet.equals(p)) {
                        p = packageInSet;
                        found = true;
                    }
                }
            } else {
                packages.add(p);
            }
            p.getEntities().add(new EntityVO(clazz.getSimpleName()));
        }
        for (final PackageVO p : packages) {
            Collections.sort(p.getEntities());
        }
        findAllNamedQueries(packages);
        return packages;
    }

    static void findAllNamedQueries(final Collection<PackageVO> annotated) throws ClassNotFoundException {
        final Iterator<PackageVO> pIt = annotated.iterator();
        while (pIt.hasNext()) {
            final PackageVO packge = pIt.next();
            final Iterator<EntityVO> qIt = packge.getEntities().iterator();
            while (qIt.hasNext()) {
                final EntityVO entity = qIt.next();
                final Class<?> clazz = Class.forName(packge.getName() + "." + entity.getName());
                final Annotation[] annotations = clazz.getAnnotations();
                for (final Annotation annotation : annotations) {
                    if (annotation.annotationType().equals(NamedQuery.class)) {
                        final NamedQuery nq = (NamedQuery) annotation;
                        entity.getQueries().add(new QueryVO(nq.name(), nq.query()));
                    } else if (annotation.annotationType().equals(NamedQueries.class)) {
                        final NamedQueries nqs = (NamedQueries) annotation;
                        for (final NamedQuery nq : nqs.value()) {
                            entity.getQueries().add(new QueryVO(nq.name(), nq.query()));
                        }
                    }
                }
            }
            Collections.sort(packge.getEntities());
        }
    }

    static Set<Class<?>> findAllEntitiesWithReflection(final Collection<String> packages) {
        final Set<Class<?>> entities = new HashSet<Class<?>>();
        for (final String packageName : packages) {
            if (packageName.split("\\.").length < 3) {
                throw new IllegalArgumentException("The package name " + packageName + " is too short, please be more specific");
            }
            final Reflections reflections = new Reflections(packageName);
            entities.addAll(reflections.getTypesAnnotatedWith(Entity.class));
            entities.addAll(reflections.getTypesAnnotatedWith(MappedSuperclass.class));
        }
        return entities;
    }
}
