/*******************************************************************************
 * Copyright 2012 pw999
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package net.sf.oqt.maven;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.persistence.Entity;
import javax.persistence.EntityManager;

import net.sf.oqt.core.CoreFactory;
import net.sf.oqt.core.CoreProperties;
import net.sf.oqt.model.EntityVO;
import net.sf.oqt.model.PackageVO;
import net.sf.oqt.model.QueryVO;
import net.sf.oqt.model.ResultVO;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.openjpa.persistence.EntityManagerImpl;
import org.apache.openjpa.persistence.OpenJPAQuery;


public abstract class AbstractQueryTranslatorTranslateMojo extends AbstractMojo {

    /**
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * The JDBC database url.
     * <p>
     * The URL to the database must be valid and the database must be up and running and acception connections. The
     * database may be empty and the user which is used to connect to this database does not required write access.
     * 
     * @parameter
     * @required
     */
    private String url = "";
    /**
     * The username for the databases connection
     * 
     * @parameter
     */
    private String username = "";
    /**
     * The password for the database connection
     * 
     * @parameter
     */
    private String password = "";
    /**
     * The drivername for the database connection.
     * <p>
     * The driver must be on the classpath. It is advised to provide the driver as a dependency for this plugin.
     * 
     * @parameter
     * @required
     */
    private String driverName = "";
    /**
     * The databse dictionary is it's known by the openJPA version on the classpath.
     * 
     * @parameter
     * @required
     */
    private String dictionary = "";
    /**
     * The list of packages where the plugin must go look for entities.
     * <p>
     * The plugin will automatically look in all the subpackages.
     * <p>
     * If this plugin is ran outside the model-module, then it's advised to give the model as a dependency of this
     * plugin.
     * 
     * @parameter
     * @required
     */
    private final List<String> packageNames = new ArrayList<String>();

    public final String getUrl() {
        return url;
    }

    public final void setUrl(final String url) {
        this.url = url;
    }

    public final String getUsername() {
        return username;
    }

    public final void setUsername(final String username) {
        this.username = username;
    }

    public final String getPassword() {
        return password;
    }

    public final void setPassword(final String password) {
        this.password = password;
    }

    public final String getDriverName() {
        return driverName;
    }

    public final void setDriverName(final String driverName) {
        this.driverName = driverName;
    }

    public final String getDictionary() {
        return dictionary;
    }

    public final void setDictionary(final String dictionary) {
        this.dictionary = dictionary;
    }

    public final List<String> getPackageNames() {
        return packageNames;
    }

    void printOutputToDebug(final ResultVO result) {
        if (getLog().isDebugEnabled()) {
            final Collection<PackageVO> packages = result.getPackages();
            for (final PackageVO packageVO : packages) {
                getLog().debug("Package: " + packageVO.getName());
                final Collection<EntityVO> entities = packageVO.getEntities();
                for (final EntityVO entityVO : entities) {
                    getLog().debug(">Entity: " + entityVO.getName());
                    final Collection<QueryVO> queries = entityVO.getQueries();
                    for (final QueryVO queryVO : queries) {
                        getLog().debug(">>NamedQuery: " + queryVO.getName());
                        getLog().debug(">>> Number of SQL queries: " + queryVO.getSQLQuery().size());
                    }
                }
            }
        }
    }

    /**
     * Checks if the required classes are on the classpath.
     * <p>
     * The checked classes are:
     * <ul>
     * <li>javax.persistence.EntityManager (javaee-api)
     * <li>javax.persistence.Entity (javaee-api)
     * <li>org.apache.openjpa.persistence.EntityManagerImpl (openJPA)
     * <li>org.apache.openjpa.persistence.OpenJPAQuery (openJPA)
     * <li>org.apache.openjpa.persistence.PersistenceProviderImpl (openJPA)
     * <li> {@link #driverName} (user provided)
     * </ul>
     * 
     * @throws MojoExecutionException if not all classes are on the classpath.
     */
    void validateClassPath() throws MojoExecutionException {
        final List<String> missing = new ArrayList<String>();

        try {
            Class.forName(EntityManager.class.getName());
            Class.forName(Entity.class.getName());
        } catch (final ClassNotFoundException e) {
            missing.add("Ensure that javaee-api is on the classpath");
        }

        try {
            Class.forName(EntityManagerImpl.class.getName());
            Class.forName(OpenJPAQuery.class.getName());
            Class.forName(org.apache.openjpa.persistence.PersistenceProviderImpl.class.getName());
        } catch (final ClassNotFoundException e) {
            missing.add("Ensure that openjpa is on the classpath");
        }

        try {
            Class.forName(getDriverName());
        } catch (final ClassNotFoundException e) {
            missing.add("Failed to find the driver" + getDriverName() + " on the classpath. Ensure that you provided it as a dependency");
        }
        if (!missing.isEmpty()) {
            for (final String string : missing) {
                getLog().error(string);
            }
            throw new MojoExecutionException("Failed to find all dependencies, see log output for more info");
        }

    }

    /**
     * Copies the properties from this plugin to the {@link CoreProperties} and verifies that the ClassPath is correct.
     */
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        // Setting properties passed via pom into the Core
        CoreFactory.getProperties().setUrl(getUrl());
        CoreFactory.getProperties().setUsername(getUsername());
        CoreFactory.getProperties().setPassword(getPassword());
        CoreFactory.getProperties().setDriverName(getDriverName());
        CoreFactory.getProperties().setDictionary(getDictionary());
        CoreFactory.getProperties().getFqns().clear();
        CoreFactory.getProperties().getPackageNames().addAll(getPackageNames());
        validateClassPath();
    }

    void setProject(final MavenProject project) {
        this.project = project;
    }

}
