/*******************************************************************************
 * Copyright 2012 pw999
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package net.sf.oqt.maven;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Locale;
import java.util.TreeSet;

import net.sf.oqt.core.Serializer;
import net.sf.oqt.model.EntityVO;
import net.sf.oqt.model.PackageVO;
import net.sf.oqt.model.QueryVO;
import net.sf.oqt.model.ResultVO;

import org.apache.maven.doxia.sink.Sink;
import org.apache.maven.doxia.sink.SinkEventAttributeSet;
import org.apache.maven.doxia.sink.SinkEventAttributes;
import org.apache.maven.doxia.siterenderer.Renderer;
import org.apache.maven.project.MavenProject;
import org.apache.maven.reporting.AbstractMavenReport;
import org.apache.maven.reporting.MavenReportException;


/**
 * The QueryTranslatorReportingMojo creates the report from the output generated
 * by the {@link #QueryTranslatorReportingMojo()}
 * 
 * @goal report
 * @execute goal="translate"
 * @phase site
 */
public class QueryTranslatorReportingMojo extends AbstractMavenReport {

    /**
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;
    /**
     * @component
     */
    private Renderer siteRenderer;
    
    /**
     * The output directory of the report.
     * @parameter default-value="${project.reporting.outputDirectory}/"
     */
    private File outputDirectory;

    /**
     * {@inheritDoc}
     */
    @Override
    public String getDescription(final Locale arg0) {
        return "The openJPA JPQL to SQL query translator report";
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getName(final Locale arg0) {
        return "Query Translator";
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getOutputName() {
        return "querytranslator";
    }

    void printOutputToDebug(final ResultVO result) {
        if (getLog().isDebugEnabled()) {
            final Collection<PackageVO> packages = result.getPackages();
            for (final PackageVO packageVO : packages) {
                getLog().debug("Package: " + packageVO.getName());
                final Collection<EntityVO> entities = packageVO.getEntities();
                for (final EntityVO entityVO : entities) {
                    getLog().debug(">Entity: " + entityVO.getName());
                    final Collection<QueryVO> queries = entityVO.getQueries();
                    for (final QueryVO queryVO : queries) {
                        getLog().debug(">>NamedQuery: " + queryVO.getName());
                        getLog().debug(">>> Number of SQL queries: " + queryVO.getSQLQuery().size());
                    }
                }
            }
        }
    }

    private void addTitleLink(final String title, final Sink sink) {
        final SinkEventAttributes attributes = new SinkEventAttributeSet();
        attributes.addAttribute(SinkEventAttributes.HREF, "#" + title.replace(" ", "_"));
        sink.link("#" + title.replace(" ", "_"), attributes);
        sink.text(title);
        sink.link_();
        sink.lineBreak();
    }

    /**
     * Generates the report based on the output of the translat mojo.
     */
    @Override
    protected void executeReport(final Locale arg0) throws MavenReportException {
        getLog().debug("Starting QueryTranslatorReportingMojo");
        try {
            final String titleEntitiesPerPackage = "Entities per packages";
            final String titleNamesPerEntity = "List of NamedQueries per entity";
            final String titleJPQLQueries = "List of JPQL queries";
            final String titleSQLQueries = "List of SQL Queries";
            final String titleFaults = "List of Faulty Queries";

            final Serializer s = new Serializer(project);
            final ResultVO result = s.deserialize();
            if (result == null) {
                getLog().warn("No result found on the output, make sure the translate goal was ran before the report goal");
                return;
            }
            if (result.getPackages().isEmpty()) {
                getLog().warn("No packages found in resulting output. Was the plugin correctly configured ? ");
            }
            
            final Collection<EntityVO> aggregatedEntities = new TreeSet<EntityVO>();
            final Collection<QueryVO> aggregatedQueries = new TreeSet<QueryVO>();
            printOutputToDebug(result);

            getLog().info("Starting creation of report");
            final Sink sink = getSink();
            sink.head();
            sink.title();
            sink.text(getName(null));
            sink.title_();
            sink.head_();

            sink.body();
            sink.section1();
            sink.sectionTitle1();
            sink.text("Report");
            sink.sectionTitle1_();

            sink.paragraph();
            addTitleLink(titleEntitiesPerPackage, sink);
            addTitleLink(titleNamesPerEntity, sink);
            addTitleLink(titleJPQLQueries, sink);
            addTitleLink(titleSQLQueries, sink);
            addTitleLink(titleFaults, sink);
            sink.paragraph_();

            sink.section2();
            sink.sectionTitle2();
            sink.text("Entities per packages");
            sink.sectionTitle2_();
            sink.table();
            sink.tableRow();
            sink.tableHeaderCell();
            sink.text("Package");
            sink.tableHeaderCell_();
            sink.tableHeaderCell();
            sink.text("Entities");
            sink.tableHeaderCell_();
            sink.tableRow_();
            for (final PackageVO packge : result.getPackages()) {
                sink.tableRow();
                sink.tableCell();
                sink.text(packge.getName());
                sink.tableCell_();
                sink.tableCell();
                sink.list();
                for (final EntityVO e : packge.getEntities()) {
                    aggregatedEntities.add(e);
                    sink.listItem();
                    sink.text(e.getName());
                    sink.listItem_();
                }
                sink.list_();
                sink.tableCell_();
                sink.tableRow_();
            }

            sink.table_();
            sink.section2_();

            sink.section2();
            sink.sectionTitle2();
            sink.text("List of NamedQueries per entity");
            sink.sectionTitle2_();
            sink.table();
            sink.tableRow();
            sink.tableHeaderCell();
            sink.text("Entity");
            sink.tableHeaderCell_();
            sink.tableHeaderCell();
            sink.text("Named Query");
            sink.tableHeaderCell_();
            sink.tableRow_();
            for (final EntityVO entity : aggregatedEntities) {
                sink.tableRow();
                sink.tableCell();
                sink.text(entity.getName());
                sink.tableCell_();
                sink.tableCell();
                sink.list();
                for (final QueryVO query : entity.getQueries()) {
                    aggregatedQueries.add(query);
                    sink.listItem();
                    sink.text(query.getName());
                    sink.listItem_();
                }
                sink.list_();
                sink.tableCell_();
                sink.tableRow_();
            }

            sink.table_();
            sink.section2_();

            sink.section2();
            sink.sectionTitle2();
            sink.text("List of JPQL queries");
            sink.sectionTitle2_();
            sink.table();
            sink.tableRow();
            sink.tableHeaderCell();
            sink.text("Name");
            sink.tableHeaderCell_();
            sink.tableHeaderCell();
            sink.text("JPQL Query");
            sink.tableHeaderCell_();
            sink.tableRow_();
            for (final QueryVO query : aggregatedQueries) {
                sink.tableRow();
                sink.tableCell();
                sink.text(query.getName());
                sink.tableCell_();
                sink.tableCell();
                sink.text(query.getJPQLQuery());
                sink.tableCell_();
                sink.tableRow_();
            }

            sink.table_();
            sink.section2_();

            sink.section2();
            sink.sectionTitle2();
            sink.text("List of SQL Queries");
            sink.sectionTitle2_();
            sink.table();
            sink.tableRow();
            sink.tableHeaderCell();
            sink.text("Name");
            sink.tableHeaderCell_();
            sink.tableHeaderCell();
            sink.text("SQL Query");
            sink.tableHeaderCell_();
            sink.tableRow_();
            for (final QueryVO query : aggregatedQueries) {
                if (query.isValid()) {
                    sink.tableRow();
                    sink.tableCell();
                    sink.text(query.getName());
                    sink.tableCell_();
                    sink.tableCell();
                    sink.list();
                    for (final String sql : query.getSQLQuery()) {
                        sink.listItem();
                        sink.text(sql);
                        sink.listItem_();
                    }
                    sink.list_();
                    sink.tableCell_();
                    sink.tableRow_();
                }
            }

            sink.table_();
            sink.section2_();

            sink.section2();
            sink.sectionTitle2();
            sink.text("List of Faulty Queries");
            sink.sectionTitle2_();
            sink.table();
            sink.tableRow();
            sink.tableHeaderCell();
            sink.text("Name");
            sink.tableHeaderCell_();
            sink.tableHeaderCell();
            sink.text("SQL Query");
            sink.tableHeaderCell_();
            sink.tableRow_();
            for (final QueryVO query : aggregatedQueries) {
                if (!query.isValid()) {
                    sink.tableRow();
                    sink.tableCell();
                    sink.text(query.getName());
                    sink.tableCell_();
                    sink.tableCell();
                    sink.text(query.getError());
                    sink.tableCell_();
                    sink.tableRow_();
                }
            }

            sink.table_();
            sink.section2_();

            sink.section1_();

            sink.body_();
            sink.flush();
            sink.close();
        } catch (ClassNotFoundException e) {
            getLog().error(e);
            throw new MavenReportException(e.getMessage());
        } catch (IOException e) {
            getLog().error(e);
            throw new MavenReportException(e.getMessage());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected MavenProject getProject() {
        return project;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Renderer getSiteRenderer() {
        return siteRenderer;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected String getOutputDirectory() {
        return outputDirectory.getAbsolutePath();
    }

}
