/*******************************************************************************
 * Copyright 2012 pw999
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package net.sf.oqt.maven;

import java.sql.SQLException;

import net.sf.oqt.core.CoreFactory;
import net.sf.oqt.core.Serializer;
import net.sf.oqt.jpa.EntityFinder;
import net.sf.oqt.jpa.JPATransformer;
import net.sf.oqt.model.ResultVO;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;


/**
 * The QueryTranslatorBuildMojo is the mojo which will convert all the known JPQL queries to SQL queries.
 * <p>
 * This plugin will gather the following information from your ORM:
 * <ul>
 * <li>The list of entities per package
 * <li>The list of named queries per entity
 * <li>The list of SQL queries per named query
 * <li>The list of erroneous queries.
 * </ul>
 * 
 * @goal translate
 * @phase site
 * @executionStrategy once-per-session
 * @author phillip
 * 
 */
public class QueryTranslatorTranslateMojo extends AbstractQueryTranslatorTranslateMojo {

    /**
     * Finds all the known entities, named queries and translates all the information.
     * <p>
     * The gathered information is saved to disk using the {@link Serializer} class.
     * 
     * @throws MojoFailureException if anything goes wrong
     * @throws MojoExecutionException if the ClassPath is not correct or if closing the database connection failed.
     */
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        getLog().info("Starting Translate Mojo");
        super.execute();
        try {
            final ResultVO result = new ResultVO();
            result.getPackages().addAll(EntityFinder.findAllEntities(CoreFactory.getProperties().getPackageNames()));
            if (result.getPackages().isEmpty()) {
                getLog().warn("Nothing found, are the packages correctly configured ? ");
            }
            CoreFactory.getProperties().setFqnsFromResult(result);
            JPATransformer.transformAllQueries(result.getPackages());
            printOutputToDebug(result);
            final Serializer s = new Serializer(project);
            s.serialize(result);

            getLog().info("Finished Translate Mojo");

        } catch (Exception e) {
            getLog().error(e);
            throw new MojoFailureException(e.getMessage());
        } finally {
            try {
                CoreFactory.shutdown();
            } catch (final SQLException e) {
                getLog().error("Failed to shutdown database connection", e);
                throw new MojoExecutionException(e.getMessage());
            }
        }

    }

}
